#ifndef TENSOR_RT_MINIMAL_H
#define TENSOR_RT_MINIMAL_H

#include <cassert>
#include <chrono>
#include <cublas_v2.h>
#include <cudnn.h>
#include <iostream>
#include <sstream>
#include <string.h>
#include <time.h>
#include <unordered_map>
#include <vector>

#include "BatchStreamPPM.h"
#include "NvInferPlugin.h"
#include "NvUffParser.h"
#include "common.h"
#include "plugin_layers.h"

using namespace nvinfer1;
using namespace nvuffparser;
using namespace plugin;

// static Logger gLogger;
static Logger gLogger(nvinfer1::ILogger::Severity::kINFO);

//INT8 Calibration, currently set to calibrate over 500 images
static constexpr int CAL_BATCH_SIZE = 50;
static constexpr int FIRST_CAL_BATCH = 0, NB_CAL_BATCHES = 10;

#define MAX_WORKSPACE (1 << 30)
#define RETURN_AND_LOG(ret, severity, message)                                              \
    do {                                                                                    \
        std::string error_message = "" + std::string(message);            \
        gLogger.log(ILogger::Severity::k ## severity, error_message.c_str());               \
        return (ret);                                                                       \
    } while(0)

inline int64_t volume(const Dims& d)
{
    int64_t v = 1;
    for (int64_t i = 0; i < d.nbDims; i++)
        v *= d.d[i];
    return v;
}

inline unsigned int elementSize(DataType t)
{
    switch (t)
    {
    case DataType::kFLOAT: return 4;
    case DataType::kHALF: return 2;
    case DataType::kINT8: return 1;
    }
    assert(0);
    return 0;
}

void* safeCudaMalloc(size_t memSize)
{
    void* deviceMem;
    CHECK(cudaMalloc(&deviceMem, memSize));
    if (deviceMem == nullptr)
    {
        std::cerr << "Out of memory" << std::endl;
        exit(1);
    }
    return deviceMem;
}

std::vector<std::pair<int64_t, DataType>>
calculateBindingBufferSizes(const ICudaEngine& engine, int nbBindings, int batchSize)
{
    std::vector<std::pair<int64_t, DataType>> sizes;
    for (int i = 0; i < nbBindings; ++i)
    {
        Dims dims = engine.getBindingDimensions(i);
        DataType dtype = engine.getBindingDataType(i);

        int64_t eltCount = volume(dims) * batchSize;
        sizes.push_back(std::make_pair(eltCount, dtype));
    }

    return sizes;
}


class RT_Inference {
    public:
        RT_Inference() {
            // FIXME: For fast-prototyping, parameters are hard coded.
            engine = NULL;
            context = NULL;

            vector<std::string> input_names;
            vector<std::string> output_names;
        }

        int create_engine(std::string fileName,
                          std::vector<string> inputs,
                          std::vector<string> outputs,
                          std::vector<DimsCHW> in_shapes) {

            auto parser = createUffParser();

            BatchStream calibrationStream(CAL_BATCH_SIZE, NB_CAL_BATCHES);

            mImageShape = in_shapes.at(0);
            parser->registerInput(inputs.at(0).c_str(), in_shapes.at(0), UffInputOrder::kNCHW);
            for (int i = 0; i < outputs.size(); ++i)
                parser->registerOutput(outputs.at(i).c_str());

            // engine = loadModelAndCreateEngine(fileName.c_str(), MAX_BATCH_SIZE, parser);
            // if (!engine) RETURN_AND_LOG(EXIT_FAILURE, ERROR, "Model load failed");
            // parser->destroy();

            // context = engine->createExecutionContext();
            // context->setProfiler(&gProfiler);


            IHostMemory* trtModelStream{nullptr};
            Int8EntropyCalibrator calibrator(calibrationStream, FIRST_CAL_BATCH, "CalibrationTableSSD");

            PluginFactory pluginFactorySerialize;
            ICudaEngine* tmpEngine = loadModelAndCreateEngine(
                fileName.c_str(), MAX_BATCH_SIZE, parser,
                &pluginFactorySerialize, &calibrator, trtModelStream);

            assert(tmpEngine != nullptr);
            assert(trtModelStream != nullptr);
            tmpEngine->destroy();
            pluginFactorySerialize.destroyPlugin();

            // parser->destroy();

            // Deserialize the engine.
            std::cout << "*** deserializing" << std::endl;
            runtime = createInferRuntime(gLogger);
            assert(runtime != nullptr);

            engine = runtime->deserializeCudaEngine(trtModelStream->data(), trtModelStream->size(), &pluginFactory);
            assert(engine != nullptr);
            trtModelStream->destroy();

            context = engine->createExecutionContext();
            assert(context != nullptr);

            // context->setProfiler(&gProfiler);

            // ------- memory allocation part
            // FIXME: Let's do this later. Allocation can be done only once.
            //        But then the allocation may require using max batch size.
            //        Trade between speed vs. memory. Not sure speed gain yet.
            // ------- paulk

            std::cout << "engine created" << std::endl;
            // execute_dummy();
            return 0;
        }


        void execute_dummy(float *dataMem)
        {
            if (!engine) {
                std::cout << "create engine first." << std::endl;
                return;
            }

            int batchSize = 1;
            int nbBindings = engine->getNbBindings();
            std::vector<void*> buffers(nbBindings);
            auto buffersSizes = calculateBindingBufferSizes(*engine, nbBindings, batchSize);

            std::cout << "batch size: " << batchSize << std::endl;
            std::cout << "nbBindings: " << nbBindings << std::endl;
            for (size_t i = 0; i < buffersSizes.size(); ++i)
                std::cout << "size of binding " << i << ": "<< buffersSizes[i].first << std::endl;

            int bindingIdxInput = 0;
            for (int i = 0; i < nbBindings; ++i)
            {
                if (engine->bindingIsInput(i)){
                    bindingIdxInput = i;
                    printf("----------input binding %d\n", i);
                }
                // else
                {
                    auto bufferSizesOutput = buffersSizes[i];
                    buffers[i] = safeCudaMalloc(bufferSizesOutput.first *
                                                elementSize(bufferSizesOutput.second));
                    printf("----------safe malloc %d, %d\n", i,
                           (int)bufferSizesOutput.first);
                }
            }

            // ----- copy input
            int eltCount = mImageShape.h() * mImageShape.w() * mImageShape.c();
            size_t memSize = eltCount * sizeof(float);
            void* deviceMem = safeCudaMalloc(batchSize * memSize);

            cudaStream_t stream;
            CHECK(cudaStreamCreate(&stream));
            CHECK(cudaMemcpyAsync(buffers[0], dataMem, batchSize * eltCount * sizeof(float), cudaMemcpyHostToDevice, stream));
            std::cout << "input image copy done." << std::endl;

            auto t_start = std::chrono::high_resolution_clock::now();
            context->execute(batchSize, &buffers[0]);
            auto t_end = std::chrono::high_resolution_clock::now();
            float ms = std::chrono::duration<float, std::milli>(t_end - t_start).count();
            std::cout << " inference takes " << ms << " ms." << std::endl;

            for (int bindingIdx = 0; bindingIdx < nbBindings; ++bindingIdx)
            {
                if (engine->bindingIsInput(bindingIdx)) continue;
                size_t memSize = buffersSizes[bindingIdx].first * elementSize(buffersSizes[bindingIdx].second);
                float *host = new float[buffersSizes[bindingIdx].first];
                CHECK(cudaMemcpy(host, buffers[bindingIdx], memSize, cudaMemcpyDeviceToHost));

                std::cout << "out, " << buffersSizes[bindingIdx].first << std::endl;

                int idx = 0;
                for (int b = 0; b < (int)batchSize; ++b)
                {
                    for (int c = 0; c < (int)(buffersSizes[bindingIdx].first / batchSize); ++c)
                    {
                        printf("%f ", host[idx++]);
                    }
                    printf("\n");
                }
                // Map<MatrixXf> v(host, (int)batchSize, (int)(buffersSizes[bindingIdx].first / batchSize));
                // outputs.push_back(v);
                delete[] host;
            }

            // ------- delete memory in device
            for (int bindingIdx = 0; bindingIdx < nbBindings; ++bindingIdx)
                // if (!engine->bindingIsInput(bindingIdx))
                CHECK(cudaFree(buffers[bindingIdx]));
            return ;
        }

        ~RT_Inference() {
            engine->destroy();
            context->destroy();
            runtime->destroy();

            pluginFactory.destroyPlugin();
        }

    public:
    private:
        ICudaEngine* engine;
        IExecutionContext* context;
        IRuntime* runtime;

        PluginFactory pluginFactory;

        void* deviceMem;
        float *dataMem;

        DimsCHW mImageShape;
        ICudaEngine* loadModelAndCreateEngine(const char* uffFile, int maxBatchSize,
                                              IUffParser* parser)
        {
            IBuilder* builder = createInferBuilder(gLogger);
            INetworkDefinition* network = builder->createNetwork();

            if (!parser->parse(uffFile, *network, nvinfer1::DataType::kFLOAT))
                RETURN_AND_LOG(nullptr, ERROR, "Fail to parse");

            /* we create the engine */
            builder->setMaxBatchSize(maxBatchSize);
            builder->setMaxWorkspaceSize(MAX_WORKSPACE);

            ICudaEngine* engine = builder->buildCudaEngine(*network);
            if (!engine)
                RETURN_AND_LOG(nullptr, ERROR, "Unable to create engine");

            /* we can clean the network and the parser */
            network->destroy();
            builder->destroy();

            return engine;
        }


        ICudaEngine* loadModelAndCreateEngine(const char* uffFile, int maxBatchSize,
                                              IUffParser* parser, nvuffparser::IPluginFactory* pluginFactory,
                                              IInt8Calibrator* calibrator, IHostMemory*& trtModelStream)
        {
            IBuilder* builder = createInferBuilder(gLogger);

            INetworkDefinition* network = builder->createNetwork();
            parser->setPluginFactory(pluginFactory);

            bool useFp16 = builder->platformHasFastFp16();
            if (useFp16)    printf("plaform has fast fp16. we will enable it.\n");
            else            printf("plaform doesn't have fast fp16.\n");

            std::cout << "Begin parsing model..." << std::endl;
            DataType modelDataType = useFp16 ? DataType::kHALF : DataType::kFLOAT; // create a 16-bit model if it's natively supported
            if (!parser->parse(uffFile, *network, modelDataType))
                RETURN_AND_LOG(nullptr, ERROR, "Fail to parse");
            std::cout << "End parsing model..." << std::endl;

            /* we create the engine */
            builder->setMaxBatchSize(maxBatchSize);
            builder->setMaxWorkspaceSize(MAX_WORKSPACE);

            if(useFp16) {
                builder->setHalf2Mode(true);
                // builder->setFp16Mode(true);
            }
            else{
                builder->setHalf2Mode(false);
            }

            std::cout << "Begin building engine..." << std::endl;
            ICudaEngine* engine = builder->buildCudaEngine(*network);
            if (!engine)
                RETURN_AND_LOG(nullptr, ERROR, "Unable to create engine");
            std::cout << "End building engine..." << std::endl;

            /* we can clean the network and the parser */
            network->destroy();
            parser->destroy();

            // Serialize the engine, then close everything down.
            trtModelStream = engine->serialize();

            builder->destroy();
            shutdownProtobufLibrary();
            return engine;
        }
};




#endif
