#ifndef __PLUGIN_LAYER_H__
#define __PLUGIN_LAYER_H__

#include <cassert>
#include <iostream>
#include <cudnn.h>
#include <cstring>

#include "NvInferPlugin.h"
#include "NvUffParser.h"

#define DEBUG_VERBOSE

using namespace nvinfer1;
using namespace nvuffparser;
using namespace plugin;


void* safeCudaMalloc_(size_t memSize)
{
    void* deviceMem;
    CHECK(cudaMalloc(&deviceMem, memSize));
    if (deviceMem == nullptr)
    {
        std::cerr << "Out of memory" << std::endl;
        exit(1);
    }
    return deviceMem;
}


void bboxTransformInvAndClip(float* rois, float* deltas, float* predBBoxes, float* imInfo,
    const int N, const int nmsMaxOut, const int numCls, int ROIS_OFFSET=1)
{
    float width, height, ctr_x, ctr_y;
    float dx, dy, dw, dh, pred_ctr_x, pred_ctr_y, pred_w, pred_h;
    float *deltas_offset, *predBBoxes_offset, *imInfo_offset;
    for (int i = 0; i < N * nmsMaxOut; ++i)
    {
        width = rois[i * (4 + ROIS_OFFSET) + 2 + ROIS_OFFSET] - rois[i * (4 + ROIS_OFFSET) + ROIS_OFFSET] + 1;
        height = rois[i * (4 + ROIS_OFFSET) + 3 + ROIS_OFFSET] - rois[i * (4 + ROIS_OFFSET) + 1 + ROIS_OFFSET] + 1;
        ctr_x = rois[i * (4 + ROIS_OFFSET) + ROIS_OFFSET] + 0.5f * width;
        ctr_y = rois[i * (4 + ROIS_OFFSET) + 1 + ROIS_OFFSET] + 0.5f * height;

        deltas_offset = deltas + i * numCls * 4;
        predBBoxes_offset = predBBoxes + i * numCls * 4;
        imInfo_offset = imInfo + i / nmsMaxOut * 3;
        for (int j = 0; j < numCls; ++j)
        {
            dx = deltas_offset[j * 4 + 0];
            dy = deltas_offset[j * 4 + 1];
            dw = deltas_offset[j * 4 + 2];
            dh = deltas_offset[j * 4 + 3];
            pred_ctr_x = dx * width + ctr_x;
            pred_ctr_y = dy * height + ctr_y;
            pred_w = exp(dw) * width;
            pred_h = exp(dh) * height;
            predBBoxes_offset[j * 4 + 0] = std::max(std::min(pred_ctr_x - 0.5f * pred_w, imInfo_offset[1] - 1.f), 0.f);
            predBBoxes_offset[j * 4 + 1] = std::max(std::min(pred_ctr_y - 0.5f * pred_h, imInfo_offset[0] - 1.f), 0.f);
            predBBoxes_offset[j * 4 + 2] = std::max(std::min(pred_ctr_x + 0.5f * pred_w, imInfo_offset[1] - 1.f), 0.f);
            predBBoxes_offset[j * 4 + 3] = std::max(std::min(pred_ctr_y + 0.5f * pred_h, imInfo_offset[0] - 1.f), 0.f);
        }
    }
}

std::vector<int> nms(std::vector<std::pair<float, int> >& score_index, float* bbox, const int classNum, const int numClasses, const float nms_threshold)
{
    auto overlap1D = [](float x1min, float x1max, float x2min, float x2max) -> float {
        if (x1min > x2min) {
            std::swap(x1min, x2min);
            std::swap(x1max, x2max);
        }
        return x1max < x2min ? 0 : std::min(x1max, x2max) - x2min;
    };
    auto computeIoU = [&overlap1D](float* bbox1, float* bbox2) -> float {
        float overlapX = overlap1D(bbox1[0], bbox1[2], bbox2[0], bbox2[2]);
        float overlapY = overlap1D(bbox1[1], bbox1[3], bbox2[1], bbox2[3]);
        float area1 = (bbox1[2] - bbox1[0]) * (bbox1[3] - bbox1[1]);
        float area2 = (bbox2[2] - bbox2[0]) * (bbox2[3] - bbox2[1]);
        float overlap2D = overlapX * overlapY;
        float u = area1 + area2 - overlap2D;
        return u == 0 ? 0 : overlap2D / u;
    };

    std::vector<int> indices;
    for (auto i : score_index)
    {
        const int idx = i.second;
        bool keep = true;
        for (unsigned k = 0; k < indices.size(); ++k)
        {
            if (keep)
            {
                const int kept_idx = indices[k];
                float overlap = computeIoU(&bbox[(idx*numClasses + classNum) * 4],
                    &bbox[(kept_idx*numClasses + classNum) * 4]);
                keep = overlap <= nms_threshold;
            }
            else
                break;
        }
        if (keep) indices.push_back(idx);
    }
    return indices;
}

//------------------------
// SliceSum layer
//------------------------
void go_kernel(float* input, float* output, int batchSize, cudaStream_t stream);

class SliceSum : public IPlugin
{
public:
    SliceSum() { std::cout << "SliceSum::Slicesum " << std::endl;};

    SliceSum(const void* data, size_t length)
    {
        std::cout << "SliceSum::Slicesum " << std::endl;
        std::cout << "length " << length << std::endl;
        const float *d = reinterpret_cast<const float*>(data);
    }
    ~SliceSum()
    {
        std::cout << "SliceSum::~Slicesum " << std::endl;
    }

    int getNbOutputs() const override { return 1; }

    Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
    {
        std::cout << "SliceSum::getOutputDimensions " << std::endl;
        std::cout << "nbInputDims" << nbInputDims << std::endl;
        std::cout << inputs[0].d[0] << inputs[0].d[1] << inputs[0].d[2]<< std::endl;
        assert(nbInputDims >= 1);
        assert(index == 0);
        return DimsCHW(2, 1, 1);
    }

    int initialize() override
    {
        return 0;
    }

    void terminate() override
    {
    }

    size_t getWorkspaceSize(int) const override { return 0; }

    int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
    {
        CHECK(cudaThreadSynchronize());

        std::cout << "SliceSum::enqueue " << std::endl;
        std::cout << "batchsize " << batchSize << std::endl;

        go_kernel((float*)inputs[0], (float*)outputs[0], batchSize, stream);

        // const float* input = reinterpret_cast<const float*>(inputs[0]);
        // float* output = reinterpret_cast<float*>(outputs[0]);

        // for (int i = 0; i < batchSize; ++i)
        // {
        //     std::cout << "b " << i << std::endl;
        //     for (int k=0; k < 12; k++)
        //         std::cout << input[k] << " ";
        //     std::cout << std::endl;
        // //     printf("%f %f\n", output[i*2+0], output[i*2+1]);
        // }
        return 0;
    }

    size_t getSerializationSize() override
    {
        return 2 * sizeof(float);
    }

    void serialize(void* buffer) override
    {
        std::cout << "SliceSum::serialize " << std::endl;
        float* d = reinterpret_cast<float*>(buffer);
        d[0] = 12.0f;
        d[1] = 24.0f;
    }

    void configure(const Dims* inputs, int nbInputs, const Dims* outputs, int nbOutputs, int) override
    {
        std::cout << "SliceSum::configure " << std::endl;
        assert(nbOutputs == 1);
    }
};


// //------------------------
// // Test layer
// //------------------------
// void run_l2_norm(float *input, float* output, int N, int H, int W, int C, cudaStream_t stream);
// class Test : public IPlugin
// {
// public:
//     MiniMatrix m;
//     Test() {
//         std::cout << "Test::Test " << std::endl;
//         m = samples_common::read_data("../../data/phantom_ssd_vgg4rt_gamma.data");
//     };

//     Test(const void* data, size_t length)
//     {
//         std::cout << "Test::Test " << std::endl;
//         std::cout << "length " << length << std::endl;
//         m = samples_common::read_data("../../data/phantom_ssd_vgg4rt_gamma.data");
//     }
//     ~Test()
//     {
//         std::cout << "Test::~Test " << std::endl;
//     }

//     int getNbOutputs() const override { return 1; }

//     Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
//     {
//         std::cout << "Test::getOutputDimensions " << std::endl;
//         std::cout << "nbInputDims" << nbInputDims << std::endl;
//         std::cout << inputs[0].d[0] << inputs[0].d[1] << inputs[0].d[2]<< std::endl;
//         assert(nbInputDims == 1);
//         assert(index == 0);
//         return inputs[0];
//         // return DimsCHW(8, 1, 1);
//     }

//     int initialize() override
//     {
//         return 0;
//     }

//     void terminate() override
//     {
//     }

//     size_t getWorkspaceSize(int) const override { return 0; }

//     int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
//     {
//         CHECK(cudaThreadSynchronize());

//         std::cout << "Test::enqueue " << std::endl;
//         std::cout << "batchsize " << batchSize << std::endl;
//         run_l2_norm((float*) inputs[0], (float*) outputs[0], 1, 1, 5, 5, stream);
//         // go_kernel((float*)inputs[0], (float*)outputs[0], batchSize, stream);
//         return 0;
//     }

//     size_t getSerializationSize() override
//     {
//         return 0;
//     }

//     void serialize(void* buffer) override
//     {
//     }

//     void configure(const Dims* inputs, int nbInputs, const Dims* outputs, int nbOutputs, int) override
//     {
//         std::cout << "Test::configure " << std::endl;
//         assert(nbOutputs == 1);
//     }
// };


void softmax(float* data, int N, int C)
{
  float *e = new float[C];
  for (int i = 0; i < N; ++i)
  {
      float max = -std::numeric_limits<float>::infinity();
      for (int c = 0; c < C; ++c){
          int idx = i*C + c;            // NxC (col-major)
          // int idx = c*N + i;            // CxN (row-major)
          e[c] = data[idx];
          if (e[c] > max) max = e[c];
      }

     float Z = 0.0;
     for (int c = 0; c < C; ++c){
        // e[i] = fasterexp(xs[i] - max);
        e[c] = exp(e[c] - max);
        Z += e[c];
     }

     for (int c = 0; c < C; ++c) {
        e[c] /= Z;                  //fmath::exp(xs[i] - max)/Z;
        int idx = i*C + c;            // NxC (col-major)
        // int idx = c*N + i;            // CxN (row-major)
        data[idx] = e[c];
    }
  }
  delete[] e;
}


/******************************/
// Flatten Plugin Layer
/******************************/
void run_flatten(float* input, float* output, int N, int C, int H, int W, cudaStream_t stream);
class Flatten : public IPlugin
{
    private:
    Dims mInShape;
    std::string mLayerName;

public:
    Flatten() {
        mLayerName = "Flatten";
#ifdef DEBUG_VERBOSE
        printf("%s::%s()\n", mLayerName.c_str(), mLayerName.c_str());
#endif
    }

    Flatten(const void* data, size_t length) {
        mLayerName = "Flatten";
#ifdef DEBUG_VERBOSE
        printf("%s::%s()\n", mLayerName.c_str(), mLayerName.c_str());
#endif
        assert(length == sizeof(int)*9);
        const int *inshape = reinterpret_cast<const int*>(data);

        int nbDims = inshape[0];
        assert(nbDims==3 || nbDims==4);
        if (nbDims == 3)
            mInShape = DimsCHW(inshape[1], inshape[2], inshape[3]);
        else
            mInShape = Dims4(inshape[1], inshape[2], inshape[3], inshape[4]);
    }

    ~Flatten() {
#ifdef DEBUG_VERBOSE
        printf("%s::~%s()\n", mLayerName.c_str(), mLayerName.c_str());
#endif
    }

    int getNbOutputs() const override { return 1; }

    Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
    {
        // assert(nbInputDims == 1);
        assert(index == 0);
        assert((inputs[0].nbDims == 3) || (inputs[0].nbDims == 4));
#ifdef DEBUG_VERBOSE
        printf("%s::getOutputDimensions()\n", mLayerName.c_str());
        printf("nbInputDims %d\n", nbInputDims);
        for (int i = 0; i < nbInputDims; ++i) {
            printf("--input %d, (", i);
            for (int j = 0; j < inputs[i].nbDims; ++j) printf("%d, ", inputs[i].d[j]);
            printf(")\n");
        }
#endif
        mInShape = inputs[0];
        return mInShape;
    }

    int initialize() override
    {
#ifdef DEBUG_VERBOSE
        printf("%s::initialize()\n", mLayerName.c_str());
#endif
        return 0;
    }

    void terminate() override
    {
#ifdef DEBUG_VERBOSE
        printf("%s::terminate()\n", mLayerName.c_str());
#endif
    }

    size_t getWorkspaceSize(int) const override
    {
#ifdef DEBUG_VERBOSE
        printf("%s::getWorkspaceSize()\n", mLayerName.c_str());
#endif
        return 0;
    }

    int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
    {
        CHECK(cudaThreadSynchronize());
        assert(mInShape.nbDims==3 || mInShape.nbDims==4);

        // printf("%s::enqueue()\n", mLayerName.c_str());
        // printf("nchw -> nhwc flattening\n");

        int N, C, H, W;
        if (mInShape.nbDims==3) {
            N = 1;
            C = mInShape.d[0];
            H = mInShape.d[1];
            W = mInShape.d[2];
        }
        else {
            N = mInShape.d[0];
            C = mInShape.d[1];
            H = mInShape.d[2];
            W = mInShape.d[3];
        }

        run_flatten((float*)inputs[0], (float*)outputs[0], N, C, H, W, stream);

        // CHECK(cudaThreadSynchronize());
        // // printf("-------------%d (%d, %d, %d, %d)\n", N*C*H*W, N, C, H, W);
        // const float *data_in = reinterpret_cast<const float*>(inputs[0]);
        // float *data_out = reinterpret_cast<float*>(outputs[0]);

        // float *local_data_in = new float[N*C*H*W];
        // float *local_data_out = new float[N*C*H*W];
        // float *local_data_out_check = new float[N*C*H*W];

        // CHECK(cudaMemcpyAsync(local_data_out_check, data_out, N*C*H*W*sizeof(float), cudaMemcpyDeviceToHost, stream));
        //     printf("kernel\n");
        // for (int k = 0; k < 100; ++k)
        //     printf("(%d, %f) ", k, local_data_out_check[k]);
        // printf("\n");


        // CHECK(cudaMemcpyAsync(local_data_in, data_in, N*C*H*W*sizeof(float), cudaMemcpyDeviceToHost, stream));
        // for (int n = 0; n < N; ++n)
        // {
        //     for (int c = 0; c < C; ++c)
        //     {
        //         for (int h = 0; h < H; ++h)
        //         {
        //             for (int w = 0; w < W; ++w)
        //             {
        //                 float v = local_data_in[n*C*H*W + c*H*W + h*W + w];
        //                 local_data_out[n*H*W*C + h*W*C + w*C +c] = v;
        //             }
        //         }
        //     }
        // }

        // printf("check\n");
        // for (int k = 0; k < N*C*H*W; ++k){
        //     if (k==94361) printf("%d: %f\n", k, local_data_in[k]);
        //     if (k==94362) printf("%d: %f\n", k, local_data_in[k]);
        //     if (k==94363) printf("%d: %f\n", k, local_data_in[k]);
        // }
        // //     if (abs(local_data_out[k]-local_data_out_check[k])>1e-3)
        // //         printf("%d: %f %f\n", k, local_data_out[k], local_data_out_check[k]);
        // // }


        // printf("\n");


        // CHECK(cudaMemcpyAsync((float*)outputs[0], local_data_out, N*C*H*W*sizeof(float), cudaMemcpyHostToDevice, stream));
        // delete[] local_data_in;
        // delete[] local_data_out;


        return 0;
    }

    size_t getSerializationSize() override
    {
#ifdef DEBUG_VERBOSE
        printf("%s::getSerializationSize()\n", mLayerName.c_str());
#endif
        return 9 * sizeof(int);
    }

    void serialize(void* buffer) override
    {
#ifdef DEBUG_VERBOSE
        printf("%s::serialize()\n", mLayerName.c_str());
#endif
        int* d = reinterpret_cast<int*>(buffer);

        d[0] = mInShape.nbDims;
        for (int i = 0; i < mInShape.nbDims; ++i)
        {
            d[i+1] = mInShape.d[i];
        }
    }

    void configure(const Dims* inputs, int nbInputs, const Dims* outputs, int nbOutputs, int) override
    {
        // assert(nbInputs == 1);
        assert(nbOutputs == 1);

#ifdef DEBUG_VERBOSE
        printf("%s::configure()\n", mLayerName.c_str());
        printf("nbInputs %d\n", nbInputs);
        for (int i = 0; i < nbInputs; ++i) {
            printf("--input %d, (", i);
            for (int j = 0; j < inputs[i].nbDims; ++j) printf("%d, ", inputs[i].d[j]);
            printf(")\n");
        }
        printf("nbOutputs %d\n", nbOutputs);
        for (int i = 0; i < nbOutputs; ++i) {
            printf("--output %d, (", i);
            for (int j = 0; j < outputs[i].nbDims; ++j) printf("%d, ", outputs[i].d[j]);
            printf(")\n");
        }
#endif
    }
};

/******************************/
// Printer Plugin Layer
/******************************/
class Printer : public IPlugin
{
public:
    std::string mLayerName;

    Printer() {
        mLayerName = "Printer";
        printf("%s::%s()\n", mLayerName.c_str(), mLayerName.c_str());
    }

    Printer(const void* data, size_t length)
    {
        mLayerName = "Printer";
        printf("%s::%s(data, length)\n", mLayerName.c_str(), mLayerName.c_str());
        std::cout << "length " << length << std::endl;
        const float *d = reinterpret_cast<const float*>(data);
    }
    ~Printer()
    {
        printf("%s::~%s()\n", mLayerName.c_str(), mLayerName.c_str());
    }

    int getNbOutputs() const override { return 1; }

    Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
    {
        assert(nbInputDims == 1);
        assert(index == 0);

        printf("%s::getOutputDimensions()\n", mLayerName.c_str());
        printf("nbInputDims %d\n", nbInputDims);
        for (int i = 0; i < nbInputDims; ++i) {
            printf("--input %d, (", i);
            for (int j = 0; j < inputs[i].nbDims; ++j) printf("%d, ", inputs[i].d[j]);
            printf(")\n");
        }

        return inputs[0];
    }

    int initialize() override
    {
        printf("%s::initialize()\n", mLayerName.c_str());
        return 0;
    }

    void terminate() override
    {
        printf("%s::terminate()\n", mLayerName.c_str());
    }

    size_t getWorkspaceSize(int) const override
    {
        printf("%s::getWorkspaceSize()\n", mLayerName.c_str());
        return 0;
    }

    int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
    {
        CHECK(cudaThreadSynchronize());

        printf("%s::enqueue()\n", mLayerName.c_str());
        std::cout << "batchsize " << batchSize << std::endl;

        // go_kernel((float*)inputs[0], (float*)outputs[0], batchSize, stream);
        return 0;
    }

    size_t getSerializationSize() override
    {
        printf("%s::getSerializationSize()\n", mLayerName.c_str());
        return 2 * sizeof(float);
    }

    void serialize(void* buffer) override
    {
        printf("%s::serialize()\n", mLayerName.c_str());
        float* d = reinterpret_cast<float*>(buffer);
        d[0] = 12.0f;
        d[1] = 24.0f;
    }

    void configure(const Dims* inputs, int nbInputs, const Dims* outputs, int nbOutputs, int) override
    {
        printf("%s::configure()\n", mLayerName.c_str());
        assert(nbOutputs == 1);
    }
};


/******************************/
// Printer Plugin Layer
/******************************/
class ImInfo : public IPlugin
{
public:
    std::string mLayerName;

    ImInfo() {
        mLayerName = "ImInfo";
        printf("%s::%s()\n", mLayerName.c_str(), mLayerName.c_str());
    }

    ImInfo(const void* data, size_t length)
    {
        mLayerName = "ImInfo";
        printf("%s::%s()\n", mLayerName.c_str(), mLayerName.c_str());
        std::cout << "length " << length << std::endl;
        const float *d = reinterpret_cast<const float*>(data);
    }
    ~ImInfo()
    {
        printf("%s::~%s()\n", mLayerName.c_str(), mLayerName.c_str());
    }

    int getNbOutputs() const override { return 1; }

    Dims getOutputDimensions(int index, const Dims* inputs, int nbInputDims) override
    {
        assert(nbInputDims == 0);
        assert(index == 0);
        return DimsCHW(3, 1, 1);
    }

    int initialize() override
    {
        printf("%s::initialize()\n", mLayerName.c_str());
        return 0;
    }

    void terminate() override
    {
        printf("%s::terminate()\n", mLayerName.c_str());
    }

    size_t getWorkspaceSize(int) const override
    {
        printf("%s::getWorkspaceSize()\n", mLayerName.c_str());
        return 0;
    }

    int enqueue(int batchSize, const void* const* inputs, void** outputs, void*, cudaStream_t stream) override
    {
        CHECK(cudaThreadSynchronize());
        printf("%s::enqueue()\n", mLayerName.c_str());
        std::cout << "batchsize " << batchSize << std::endl;
        float im_info[3] = {288., 512., 1.0};
        CHECK(cudaMemcpyAsync(reinterpret_cast<float*>(outputs[0]),
                              im_info,
                              3*sizeof(float), cudaMemcpyHostToDevice, stream));
        return 0;
    }

    size_t getSerializationSize() override
    {
        printf("%s::getSerializationSize()\n", mLayerName.c_str());
        return 0;
    }

    void serialize(void* buffer) override
    {
        printf("%s::serialize()\n", mLayerName.c_str());
    }

    void configure(const Dims* inputs, int nbInputs, const Dims* outputs, int nbOutputs, int) override
    {
        printf("%s::configure()\n", mLayerName.c_str());
        assert(nbInputs == 0);
        assert(nbOutputs == 1);
    }
};


// Integration for serialization.
class PluginFactory : public nvinfer1::IPluginFactory, public nvuffparser::IPluginFactory
{
public:
    virtual nvinfer1::IPlugin* createPlugin(const char* layerName, const nvinfer1::Weights* weights, int nbWeights, const nvuffparser::FieldCollection fc) override
    {
        assert(isPlugin(layerName));

        // const nvuffparser::FieldMap* fields = fc.fields;
        // int nbFields = fc.nbFields;
        std::cout<< layerName << std::endl;
        if(!strcmp(layerName, "_slicesum"))
        {
            assert(mPluginSliceSum.get() == nullptr);
            mPluginSliceSum = std::unique_ptr<SliceSum>(new SliceSum());
            return mPluginSliceSum.get();
        }
        if (!strcmp(layerName, "_Printer"))
        {
            assert(mPluginPrinter.get() == nullptr);
            mPluginPrinter = std::unique_ptr<Printer>(new Printer());
            return mPluginPrinter.get();
        }
        if (!strcmp(layerName, "_flatten"))
        {
            // assert(mPluginFlatten.get() == nullptr);
            mPluginFlatten = std::unique_ptr<Flatten>(new Flatten());
            return mPluginFlatten.get();
        }
        else
        {
            assert(0);
            return nullptr;
        }
    }

    IPlugin* createPlugin(const char* layerName, const void* serialData, size_t serialLength) override
    {
        std::cout<< layerName << std::endl;
        assert(isPlugin(layerName));

        if (!strcmp(layerName, "_slicesum") || !strncmp(layerName, "_slicesum", 9))
        {
            assert(mPluginSliceSum.get() == nullptr);
            mPluginSliceSum = std::unique_ptr<SliceSum>(new SliceSum(serialData, serialLength));
            return mPluginSliceSum.get();
        }
        // if (!strcmp(layerName, "_test"))
        // {
        //     assert(mPluginTest.get() == nullptr);
        //     mPluginTest = std::unique_ptr<Test>(new Test(serialData, serialLength));
        //     return mPluginTest.get();
        // }
        if (!strcmp(layerName, "_Printer"))
        {
            assert(mPluginPrinter == nullptr);
            mPluginPrinter = std::unique_ptr<Printer>(new Printer(serialData, serialLength));
            return mPluginPrinter.get();
        }
        if (!strcmp(layerName, "_flatten") || !strncmp(layerName, "_flatten", 8))
        {
            assert(mPluginFlatten == nullptr);
            mPluginFlatten = std::unique_ptr<Flatten>(new Flatten(serialData, serialLength));
            return mPluginFlatten.get();
        }
        else
        {
            assert(0);
            return nullptr;
        }
    }

    bool isPlugin(const char* name) override
    {
        return (!strcmp(name, "_slicesum")
             || !strcmp(name, "_test")
             || !strcmp(name, "_flatten")
             || !strncmp(name, "_flatten", 8)
             || !strcmp(name, "_Printer"));
    }

    // The application has to destroy the plugin when it knows it's safe to do so.
    void destroyPlugin()
    {
        mPluginSliceSum.reset();
        mPluginSliceSum = nullptr;

        mPluginPrinter.reset();
        mPluginPrinter = nullptr;

        mPluginFlatten.reset();
        mPluginFlatten = nullptr;
    }

    std::unique_ptr<SliceSum> mPluginSliceSum{nullptr};
    std::unique_ptr<Printer> mPluginPrinter{nullptr};
    std::unique_ptr<Flatten> mPluginFlatten{nullptr};
};




#endif
